//------------------------------------------------------------------------------
// Цифровой модуль
// Список изменений - в конце файла
// OldBean, 09.02.2018
//------------------------------------------------------------------------------
#include <avr/io.h>
#include <util/twi.h>
#include <avr/interrupt.h>

#define F_CPU 16000000UL
#include <util/delay.h>

#define ADDR 0x15  // Адрес устройства на шине I2C
#define MAX_BUF 16 // Размер буфера обмена    

void dout(uint8_t byte);
void owReset();
void owWrite(uint8_t b);
void owRead(uint8_t *b);

static volatile uint8_t index = 0;    // Номер отправляемого байта
static volatile uint8_t buf[MAX_BUF]; // Буфер обмена

static volatile uint8_t owdevs; // На пинах где обнаружено 1-Wire-устройство - 1
static uint8_t bytes[8];
//------------------------------------------------------------------------------
// Функция манипулирует состоянием порта D, к каждому пину которого подключены
// однопроводные шины, организованные по "схемному И" и притянуты к +5В внешними
// резисторами. 
// Если бит входного параметра (byte) равен 0, то соответствующий пин порта D
// притягивается к нулю. Если же бит равен 1, то соответствующий пин 
// переводится в Z-состояние.
//------------------------------------------------------------------------------
void dout(uint8_t byte) {
  DDRD = ~byte;  // Где нулевые биты - там на вывод, где единички - на ввод
  PORTD &= byte; // Притягиваем пины, где нулевые биты, к земле
}
//------------------------------------------------------------------------------
ISR(TWI_vect) { // Обработчик событий модуля TWI
  uint16_t w;
  switch(TW_STATUS) {      // Статус - в 5 старших битах регистра TWSR
    case TW_SR_DATA_ACK:   // Данные приняты, ACK отправлен - 0x80
      index = TWDR << 1;   // На младший байт запрашиваемого слова
      if(index >= MAX_BUF) index = 0;
      break;
    case TW_ST_SLA_ACK:    // Запрос SLA+R, ACK отправлен - 0xA8
      TWDR = buf[index++]; // Очередной байт - в регистр данных TWI
      if(index >= MAX_BUF) index = 0; // Выбрали буфер - вернемся к его началу
      break;               // последовательных байта (начиная с младшего)
    case TW_BUS_ERROR:     // Неверные условия "старт" или "стоп" - 0x00
      TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN); // "Сброс" TWI
      break;
  }
  TWCR |= _BV(TWINT); // Флаг TWINT здесь НЕ очищается автоматически! Очистим.
} 
//------------------------------------------------------------------------------
// Таймер 0 используется для "долговременного" тактирования 1 сек и 752 мс (это
// с небольшим запасом на reset и запуск преобразования). 
//------------------------------------------------------------------------------
ISR(TIMER0_OVF_vect) { // Обработчик по переполнению
  volatile static uint8_t c1000 = 0, c0750 = 0;
  if(++c1000 == 61) { // Сброс и запуск преобразования
    c1000 = 0; c0750 = 0;
    sei(); // Разрешаем прерывания для (некритичной) работы с 1-Wire
    owReset();
    owWrite(0xcc); // Skip ROM
    owWrite(0x44); // Запускаем все датчики
  }
  if(++c0750 == 47) { // Преобразование закончено, данные готовы - читаем их
    sei(); // Разрешаем прерывания для (некритичной) работы с 1-Wire
    owReset();
    owWrite(0xcc); // Skip ROM
    owWrite(0xbe); // Будем читать данные
    owRead(bytes);
    for(uint8_t i = 0; i < 8; i++)
      buf[1 + 2*i] = bytes[i];
    owRead(bytes);
    for(uint8_t i = 0; i < 8; i++)
      buf[2*i] = bytes[i];
    owReset();
  }
}
//------------------------------------------------------------------------------
void owReset() { // Сброс и обнаружение подключенных к шинам 1-Wire датчиков
  dout(0x00); _delay_us(480); // Сброс всех линий
  dout(0xff); _delay_us(60); // Отпустим линии и выдержим паузу
  owdevs = ~PIND; // Это не имеет сойчас особого смысла - на всякий случай
  _delay_us(420); //ждем оставшееся время до 480мкс
}
//-----------------------------------------------------------------------------------------------------------
void owWrite(uint8_t b) { // Отсылаем один и тот же байт на все шины
  uint8_t mask = 0b00000001;
  for(uint8_t bit = 0; bit < 8; bit++) {
    dout(0x00); _delay_us(1);  // Начало слота: все линии - в ноль
    if(b & mask) dout(0xff); // Отпустим линии, если нужно передать единичку
    _delay_us(59);
    dout(0xff); // Все равно отпускаем линии в конце слота
    _delay_us(5);
    mask <<= 1;
  }
}
//-----------------------------------------------------------------------------
void owRead(uint8_t *b) { // Чтение байтов со всех линий
  uint8_t wb[8]; // Вспомогательный массив, который заносятся читаемые биты.
                 // Каждый байт wb содержит соответствующие биты всех датчиков.
                 // Например, wb[3] содержит 3-и биты всех датчиков.
  for(uint8_t bit = 0; bit < 8; bit++) { // Пробегаем 
    dout(0x00); _delay_us(1);  // Начало слота: все линии - в ноль
    dout(0xff); _delay_us(14); // Отпустим все линии и пауза
    wb[bit] = PIND; // Читаем со всех линий биты с номером bit
    _delay_us(45); // Пауза до конца слота
  }
  _delay_us(5);
  // Преобразуем полученную информацию (массив wb) к нормальной форме, чтобы
  // каждый байт содержал все биты, полученные от одного датчика. По сути,
  // нам нужно просто транспонировать матрицу 8x8, представленную массивом wb.
  uint8_t bmask = 0b00000001;
  for(uint8_t bit = 0; bit < 8; bit++) {
    uint8_t nmask = 0b00000001;
    for(uint8_t sn = 0; sn < 8; sn++) {
      if(wb[bit] & nmask) b[sn] |= bmask; // Установим бит
      else b[sn] &= ~bmask; // Сбросим бит
      nmask <<= 1;
    }
    bmask <<= 1;
  }
}
//------------------------------------------------------------------------------
int main(void) {
  MCUCR |= _BV(PUD);  // Отключаем подтягивающие резисторы на всех портах
  // Инициализация TWI-модуля --------------------------------------------------
  TWAR = ADDR << 1;   // Адрес устройства - в старших 7 битах
  TWCR = _BV(TWIE);   // Разрешаем прерывания от TWI
  TWCR |= _BV(TWEA);  // Разрешаем подачу импульса ACK
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
  TWCR |= _BV(TWEN);  // Разрешаем работу TWI и активирует TWI-интерфейс
  // Инициализация таймера 0 для долговременного тактирования ------------------
  // Счетчик работает в обычном режиме (Normal mode)
  TCCR0B = _BV(CS02) | _BV(CS00); // Предделитель на 1024
  TIMSK0 |= _BV(TOIE0); // Разрешаем прерывания по переполнению

  sei(); // Разрешаем глобальные прерывания
  while(1) { // Бесконечный цикл
  }
  return 0;
}
//------------------------------------------------------------------------------
// Список изменений и история
// 1. 09.02.2018 - Шаблон прошивки - проверка работоспособности модуля
// 2. 20.02.2018 - Отлажен базовый вариант цифрового модуля для двух типов 
//                 датчиков: DS18B20 и контакты
// 3. 21.02.2018 - Разрешены глобальные прерывания в периоды работы с 1-Wire
//------------------------------------------------------------------------------

